---------------------------------------------------------------------
------------------------ All freeslot Extra  ------------------------
---------------------------------------------------------------------


--Sprites
freeslot("SPR2_CAR_", "SPR2_TFCA", "SPR2_TFBO", "SPR_BLST", "SPR_BULL", "SPR2_BUSN", "SPR2_BUJU")
--States
freeslot("S_CAR_MODE",
"S_TF_TO_BOT", "S_TF_TO_CAR", "S_BLAST_SHOT", "S_BULL_SHOT", "S_BULLET_STANCE")

--Sounds
freeslot("sfx_tf_car", "sfx_tf_bot", "sfx_engine", "sfx_brake", "sfx_bump", "sfx_bullet")

--Objects
freeslot("MT_BLAST", "MT_BULL")


------------------------------------------------------------------------------------
------------------------ All rawset Variables and Functions ------------------------
------------------------------------------------------------------------------------
--- NOTE: Constant variables are FULLY CAPITALIZED (don't change constants!!!) -----
------------------------------------------------------------------------------------


--Car constants
rawset(_G, "CAR_ACCELERATION", 4*FRACUNIT) --Force applied every tic
rawset(_G, "SPEED_LIMIT", 65*FRACUNIT) --Fastest speed whlie driving
rawset(_G, "STEERING_LIMIT_ANGLE", {MAX = ANGLE_45, MIN = -ANGLE_45}) --Furthest steering angle
rawset(_G, "REGULAR_STEERING_SPEED", FRACUNIT/13) --How fast can you steer regularly
rawset(_G, "DRIFTING_STEERING_SPEED", FRACUNIT/8) --How fast can you steer while drifting
rawset(_G, "BUMP_MF", MF_ENEMY)
rawset(_G, "BUMP_NOT_MF", 0) 
rawset(_G, "BULLET_TICS_MAX", TICRATE/3) --no clude
rawset(_G, "TRANFORMATION_TICS_MAX", TICRATE/4) --The cooldown of the transformation
rawset(_G, "TO_BOT_JUMP", FRACUNIT*10) --Jump force applied whlie transforming into bot while holding forwatd drive button


-- Simulates Enums for all possible transformation modes
rawset(_G, "TF_MODES", {BOT_MODE = 0, CAR_MODE = 1})

--Clamps the integer num between inlcusive bounds
rawset(_G, "clamp", function(num, lower, upper)
	if(num > upper) then
		num = upper
	elseif(num < lower) then
		num = lower
	end
	return num
	-- return max(lower, min(upper, num))
end)



--Checks whether the mobject is valid and (optionally) has the correct skin 
rawset(_G, "Valid", function(mo, skin)
	return mo ~= nil and mo.valid == true and mo.skin == skin and mo.state ~= S_NULL and mo.player.playerstate == PST_LIVE--and mo.state ~= states[mo.state].deathstate
end)


--Function mostly shown by clairebun to use on SRB2 discord page
--mo1 (mobj_t): first mobj to check for collision
--mo2 (mobj_t): second mobj to check for collision
--extraheight (int): extra height added to the mo2
--returns: true if two mobjects collide vertically, false otherwise
rawset(_G, "L_ZCollide", function(mo1, mo2, extraheight)
	if(extraheight == nil) then
		extraheight = 0
	end
    if mo1.z > mo2.height+mo2.z+extraheight then return false end
    if mo2.z > mo1.height+mo1.z then return false end
    return true
end)

--Spawns the objects on the side of the player
rawset(_G, "SpawnOnSide", function(player, toSpawn)
    local right_x = player.mo.x + FixedMul(skins[player.mo.skin].radius, InvAngle(cos(player.mo.angle+ANGLE_90)))
    local right_y = player.mo.y + FixedMul(skins[player.mo.skin].radius, InvAngle(sin(player.mo.angle+ANGLE_90)))
    local left_x  = player.mo.x + FixedMul(skins[player.mo.skin].radius, InvAngle(cos(player.mo.angle-ANGLE_90)))
    local left_y  = player.mo.y + FixedMul(skins[player.mo.skin].radius, InvAngle(sin(player.mo.angle-ANGLE_90)))
    
    --Spawn at the side
    P_SpawnMobj(right_x, right_y, player.mo.z, toSpawn)
    P_SpawnMobj(left_x, left_y, player.mo.z, toSpawn)
end)


--Switch to the jazz skin
--Returns true if could switch to jazz, false if player was already it
rawset(_G, "SwitchTo", function(player)
	if(Valid(player.mo)) then
		return false -- do not switch to jazz if the player is already jazz
	end
    
	--Set controls
	player.jumpheld = 0
	player.spinheld = 0
    player.forwardheld = 0
	player.prevjumpheld = 0
	player.prevspinheld = 0
	
    --The previous vertical momentum
    player.mo.prevmomz = 0
	
    --Is jump special available?
    player.jumpspecial = 0 -- true if releasing jump in the air, enable implementing jumping special abilities manual
    
    --The current transformation mode of jazz
    if(player.tf_mode == nil) then
        player.tf_mode = TF_MODES.BOT_MODE -- The default jazz mode is botmode
    end
    player.transformation_tics = TRANFORMATION_TICS_MAX --Transformation is allowed only when tics <= 0 (so this is its cooldown)
    player.bullet_tics = 0 --No clue
    
    --The angle of car steering
    player.mo.steering = 0 
    player.mo.applied_speed = 0
    player.mo.applied_angle = 0
    player.forward_input = 0
    player.braking_input = 0

    -- True if the jazz was on the ground in the previous frame
    player.mo.prevcarground = true

    return true
end)


--Switch off the jazz skin
--Returns true if could switch offset jazz, false if player wan't jazz in the first place
rawset(_G, "SwichOff", function(player)
	if(Valid(player.mo)) then
		return false
	end


    if(player.tf_mode == TF_MODES.CAR_MODE) then
        player.mo.state = S_TF_TO_BOT
    end

    --Remove controls
	player.jumpheld = nil
	player.spinheld = nil
    player.forwardheld = nil
	player.prevjumpheld = nil
	player.prevspinheld = nil

	player.mo.prevmomz = nil
	player.jumpspecial = nil

    player.tf_mode = nil
    player.transformation_tics = nil
    
    player.mo.steering = nil

    player.mo.applied_speed = nil
    player.mo.applied_angle = nil
    player.forward_input = nil
    player.braking_input = nil

    player.mo.prevcarground = nil
    player.mo.prevskin = nil

    return true
end)



------------------------------------------------------------
------------------ All Sound definitions -------------------
------------------------------------------------------------



-- Transformion to car
sfxinfo[sfx_tf_car] = {
	singular = false,
	priority = 64
}

-- Transformion to bot
sfxinfo[sfx_tf_bot] = {
	singular = false,
	priority = 64
}
-- Engine sound
sfxinfo[sfx_engine] = {
	singular = false,
	priority = 65
}
-- Braking sound
sfxinfo[sfx_brake] = {
	singular = false,
	priority = 64
}
-- Braking sound
sfxinfo[sfx_bump] = {
	singular = false,
	priority = 64
}
-- Bullet #1 sound
sfxinfo[sfx_bullet] = {
	singular = false,
	priority = 64
}



--------------------------------------------------------------------
------------------------ Important Thinkers ------------------------
--------------------------------------------------------------------


--The Base Thinker that plays before others,
--Mostly used to record players input
addHook("PreThinkFrame", function()
	for player in players.iterate() do
        
        --Skin switching behavior
        if(player.mo.prevskin == "jazz" and player.mo.prevskin ~= player.mo.skin) then
            SwichOff(player)
        elseif(player.mo.skin == "jazz" and player.mo.prevskin == nil) then
            SwitchTo(player)
        end
        
        --Prevent any non-jazz players to execute behavior afte rit
        -- continue -- Try find the next jazz player if current is not jazz
        if(player.mo.skin == "jazz") then
        
            -- Prevent drowning in liquid and suffocating in space
            player.powers[pw_underwater] = 0
            player.powers[pw_spacetime] = 0

            -- Controls 
            --Jump button
            if(player.cmd.buttons & BT_JUMP) then
                player.jumpheld = $+1
            else
                player.jumpheld = 0
            end

            --Spin button
            if(player.cmd.buttons & BT_SPIN) then
                player.spinheld = $+1
            else
                player.spinheld = 0
            end

            --Forward button (accounting for joystick too)
            if(player.cmd.forwardmove >= 40) then
                player.forwardheld = $+1
            else
                player.forwardheld = 0
            end

            --Jump special ability (available when letting go of jump button in the air)
            if(not P_IsObjectOnGround(player.mo) and player.jumpheld == 0) then
                player.jumpspecial = 1
            elseif(player.mo.eflags & MFE_JUSTHITFLOOR or player.powers[pw_carry] ~= 0) then
                player.jumpspecial = 0
            end
        end
        
    end
end)



-- Regular behaviors for both modes (such as Passive Skills)
addHook("PlayerThink", function(player)
    
    if(not Valid(player.mo, "jazz")) then
        return 
    end
    
end)



--The Thinker that plays after other thikers,
--mostly used to clean up, record the previous state, 
--and jump and spin button holding
addHook("PostThinkFrame", function()
	for player in players.iterate() do
		if(Valid(player.mo, "jazz")) then

            -- Prevent changing states
            -- if (player.mo.state ~= S_CAR_MODE and player.mo.state ~= S_TF_TO_BOT) then 
            if(player.tf_mode == TF_MODES.CAR_MODE and player.mo.state ~= S_TF_TO_BOT and player.mo.state ~= S_TF_TO_CAR) then
                player.mo.state = S_CAR_MODE
            end

			player.mo.prevskin = player.mo.skin

			player.prevjumpheld = player.jumpheld
			player.prevspinheld = player.spinheld


            --Recording if the car was on the ground in previous tic
            player.mo.prevcarground = P_IsObjectOnGround(player.mo)

		end
	end
end)



-------------------------------------------------------------
------------------------ All Actions ------------------------
-------------------------------------------------------------

local function A_BULLET_STANCE(actor, var1, var2)
    if(not Valid(actor, "jazz")) then
        return false --Wasn't called for jazz
    end

    actor.player.bullet_tics = BULLET_TICS_MAX
end

-- Sets up jazz for transformation into the bot
local function A_TF_TO_BOT(actor, var1, var2)
    -- print("to bot!")
    -- if(not Valid(actor, "jazz")) then
    --     return false --Wasn't called for jazz
    -- end
    --Restore mouse input
    input.setMouseGrab(true)

    -- Restore the original height
    actor.player.height = skins[actor.player.skin].height

    -- Change the current transformation state
    actor.player.tf_mode = TF_MODES.BOT_MODE
    actor.player.transformation_tics = TRANFORMATION_TICS_MAX -- resetting transformation tics

    --Breaking floors
    actor.player.powers[pw_strong] = $|STR_FLOOR

    actor.player.charflags = $&~SF_NOSKID

    --Has jazz jumped while transforming?
    local jumped = false
    -- Artificial jump while transforming 
    if(actor.player.jumpheld >= 1 or input.gameControlDown(GC_FIRE) ~= 0 and P_IsObjectOnGround(actor)) then
        --Actual jump force applied
        local actual_jump = TO_BOT_JUMP
        if(actor.eflags & MFE_UNDERWATER ~= 0) then
            --Jump underwater is redued
            actual_jump = $/2
        end
        P_SetObjectMomZ(actor, actual_jump, true)

        jumped = true
    end

    --Playing correct next animations after transformatino!
    if(jumped or not P_IsObjectOnGround(actor)) then
        --Play jumping animation
        states[actor.state].nextstate = S_PLAY_JUMP
    else
        --Play normal animations
        states[actor.state].nextstate = S_PLAY_STND
    end
    
    -- Play sound
    S_StartSound(actor, sfx_tf_bot)
end


-- Sets up jazz for transformation into the car
local function A_TF_TO_CAR(actor, var1, var2)
    if(not Valid(actor, "jazz")) then
        return false --Wasn't called for jazz
    end

    --Ignore mouse input
    input.setMouseGrab(false)

    -- Reduce the height of car to fit into the spingap
    actor.player.height = actor.player.spinheight

    --Stop breaking floors
    actor.player.powers[pw_strong] = $&~STR_FLOOR

    actor.player.charflags = $|SF_NOSKID

    -- Change the current transformation state
    actor.player.tf_mode = TF_MODES.CAR_MODE
    actor.player.transformation_tics = TRANFORMATION_TICS_MAX -- resetting transformation tics

    -- Play sound
    S_StartSound(actor, sfx_tf_car)
end



------------------------------------------------------------
------------------ All State definitions -------------------
------------------------------------------------------------


--State of jet
states[S_BULLET_STANCE] = {
	sprite = SPR_PLAY,
	frame = SPR2_BUSN,
	tics = TICRATE,
    action = A_BULLET_STANCE,
    nextstate = S_PLAY_STND
}

--State of jet
states[S_CAR_MODE] = {
	sprite = SPR_PLAY,
	frame = SPR2_CAR_,
	tics = -1
}

-- Transformation sequence into the car mode
states[S_TF_TO_CAR] = {
	sprite = SPR_PLAY,
	frame = SPR2_TFCA|FF_ANIMATE|A,
	tics = 2*(TICRATE/5),
	action = A_TF_TO_CAR,
    var1 = 2,
    var2 = TICRATE/5,
    nextstate = S_CAR_MODE
}

-- Transformation sequence into the bot mode
states[S_TF_TO_BOT] = {
	sprite = SPR_PLAY,
	frame = SPR2_TFBO|FF_ANIMATE|A,
	tics = states[S_TF_TO_CAR].tics,
	action = A_TF_TO_BOT,
    var1 = states[S_TF_TO_CAR].var1,
    var2 = states[S_TF_TO_CAR].var2,
	nextstate = S_PLAY_STND
}



-- Projectile: Blast that is shot
states[S_BLAST_SHOT] = {
	sprite = SPR_BLST,
    frame = A,
	tics = -1
}


-- Projectile: Bullet that is shot
states[S_BULL_SHOT] = {
	sprite = SPR_BULL,
    frame = A,
	tics = TICRATE
}



------------------------------------------------------------
----------------- All Objects definitions ------------------
------------------------------------------------------------




mobjinfo[MT_BLAST] = {
    spawnstate = S_BLAST_SHOT,
    speed = FRACUNIT,
    radius = 32*FRACUNIT,
    height = 32*FRACUNIT,
    damage = 2,
    flags = MF_MISSILE|MF_GRENADEBOUNCE|MF_NOGRAVITY
}



mobjinfo[MT_BULL] = {
    spawnstate = S_BULL_SHOT,
    speed = 100*FRACUNIT,
    radius = 10*FRACUNIT,
    height = 10*FRACUNIT,
    damage = 1,
    flags = MF_MISSILE|MF_NOGRAVITY
}